/*
* Copyright(C) 2004 Marcelo Leal.
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.

* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.

* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <pwd.h>
#include <syslog.h>
#include <security/pam_appl.h>
#include <security/pam_misc.h>
#include <security/_pam_types.h>
#include <lber.h>
#include <ldap.h>

extern int h_errno;
#define PAM_MODULE_NAME  "pam_hostscheck"
#define PAM_HOSTSCHECK_TRUSTED_FILE "/etc/pam_hostscheck-users.conf"
#define PAM_SM_AUTH

/* --- authentication management functions (the game :) --- */

PAM_EXTERN
int pam_sm_authenticate(pam_handle_t *pamh, int flags, int argc, const char **argv) {
 int cod_retorno, ln, fhr, length;
 struct passwd *pwd;
 LDAP *ld;
 LDAPMessage *ldap_message_set;
 char *user, *ldap_host, *base_dn, *trusted_file, *trusted_user;
 char filter[512] = "(&(uid=";
 char myhostname[256], buffer[256], *attributes[2] = {"dn", NULL};

 // Arguments...
 ldap_host = argv[0];
 base_dn = argv[1];
 syslog(LOG_INFO, "pam_hostscheck: LDAP Server: %s", ldap_host);
 syslog(LOG_INFO, "pam_hostscheck: Base DN: %s", base_dn);

 // The host name...
 if (gethostname(myhostname, 255) != 0) {
  syslog(LOG_ERR, "pam_hostscheck: Couldn't determine my hostname.");
  return PAM_SYSTEM_ERR;
 } 
 syslog(LOG_INFO, "pam_hostscheck: host: %s.", myhostname);

 // The user name...
 cod_retorno = pam_get_item(pamh, PAM_USER, (const void **) &user);
 if (cod_retorno != PAM_SUCCESS || user == NULL || *user == '\0') {
   syslog(LOG_ERR, "pam_hostscheck: Couldn't identify user %s.", pwd->pw_name);
   return PAM_SYSTEM_ERR;
  }
 pwd = getpwnam(user);
 if (pwd == NULL) {
  syslog(LOG_ERR, "pam_hostscheck: user: %s unknown.", pwd->pw_name);
  return PAM_USER_UNKNOWN;
 }
 // The One...
 if (pwd->pw_uid == 0) {
    syslog(LOG_INFO, "pam_hostscheck: You are the One!");
    return PAM_SUCCESS;
 }
 // Read a trusted file... 
 trusted_file = PAM_HOSTSCHECK_TRUSTED_FILE;
 if (!(fhr = fopen(trusted_file, "r"))) {
  syslog(LOG_ERR, "pam_hostscheck: Warning: We are working without a trusted file, only LDAP users can authenticate.");
  } else {  
  // parser by branden.. :))
  ln = 0;
  while (fgets(buffer, sizeof(buffer), fhr)) {
   ln++;
   length = strlen(buffer);
   if (buffer[length - 1] != '\n') {
    syslog(LOG_WARNING, "pam_hostscheck: line too long at %s:%d, ignored", trusted_file, ln); 
    while (fgets(buffer, sizeof(buffer), fhr)) {
     if (buffer[strlen(buffer) - 1] == '\n') break;
    }
    continue;
   }
   buffer[length - 1] = '\0';
   for (trusted_user = buffer; *trusted_user == ' ' || *trusted_user == '\t'; trusted_user++);
   if (*trusted_user == '#' || *trusted_user == '\0') continue;
   if (strcmp (trusted_user, pwd->pw_name) == 0 ) {
    syslog(LOG_INFO, "pam_hostscheck: user: %s is welcome here!", trusted_user);
    return PAM_SUCCESS;
   }
  }
 }
 /* I'm sure that there is a better way to do this.. :) */
 strcat (filter, pwd->pw_name);
 strcat (filter, ")");
 strcat (filter, " (|(host=todos_os_servidores) (host=");
 strcat (filter, myhostname);
 strcat (filter, ")))");
 syslog(LOG_INFO, "pam_hostscheck: filter: %s.", filter);
 if ((ld = ldap_init(ldap_host, LDAP_PORT)) == NULL ) {
  syslog(LOG_ERR, "pam_hostscheck: ldap_init error!");
  return PAM_SYSTEM_ERR;  
 }
 cod_retorno = ldap_search_s (ld, base_dn, LDAP_SCOPE_SUBTREE, filter, attributes, 0, &ldap_message_set);
 if (cod_retorno != LDAP_SUCCESS ) {
  ldap_perror (ld, "ERROR in ldap_search_s");
  return PAM_SYSTEM_ERR;  
 } 
 if (ldap_count_entries(ld, ldap_message_set) == 0) {
  syslog(LOG_ERR, "pam_hostscheck: user: %s not allowed on this host!", pwd->pw_name);
  cod_retorno = PAM_AUTH_ERR; 
  } else {
  syslog(LOG_INFO, "pam_hostscheck: user: %s OK on this host!", pwd->pw_name);
  cod_retorno = PAM_SUCCESS;
 }
 return cod_retorno;
}

PAM_EXTERN
int pam_sm_setcred(pam_handle_t *pamh, int flags, int argc, const char **argv) {
    // just one hack... :)
    return PAM_SUCCESS;
}


#ifdef PAM_STATIC

/* static module data */
struct pam_module _pam_hostscheck_modstruct =
{
   "pam_hostscheck",
   NULL,
   NULL,
   NULL,
   pam_sm_authenticate,
   pam_sm_setcred,
   NULL,
};

#endif
